@@ -7,6 +7,7 @@
 
 #include "history.h"
 #include "keymap.h"
+#include "lineedit.h"
 #include "lineview.h"
 #include "listwidget.h"
 #include "stflpp.h"
@@ -33,12 +34,6 @@ enum class CommandType {
 	INVALID, 	/// differs from UNKNOWN in that no input was parsed
 };
 
-enum class BindingType {
-	BindKey,
-	Macro,
-	Bind,
-};
-
 struct Command {
 	CommandType type;
 	std::vector<std::string> args;
@@ -49,6 +44,18 @@ struct Command {
 	{}
 };
 
+enum class QnaFinishAction {
+	None,
+	Bookmark,
+	RunCmdLine,
+	GotoUrlByNumber,
+	UpdateFlags,
+	Search,
+	PipeItemIntoProgram,
+	GotoTitle,
+	SetFilter,
+};
+
 class FormAction {
 public:
 	FormAction(View&, std::string formstr, ConfigContainer* cfg);
@@ -78,17 +85,23 @@ public:
 		const std::vector<std::string>& args,
 		BindingType bindingType = BindingType::BindKey);
 
-	virtual void finished_qna(Operation op);
+	virtual void finished_qna(QnaFinishAction op);
 
 	void start_cmdline(std::string default_value = "");
 
 	void start_qna(const std::vector<QnaPair>& prompts,
-		Operation finish_op,
+		QnaFinishAction finish_op,
 		History* h = nullptr);
 	void finish_qna_question();
 	void cancel_qna();
 	void qna_next_history();
 	void qna_previous_history();
+	void clear_line();
+	void clear_eol();
+	void delete_word();
+	void handle_cmdline_completion();
+
+	void handle_qna_event(std::string event, bool inside_cmd);
 
 	void set_parent_formaction(std::shared_ptr<FormAction> fa)
 	{
@@ -163,10 +176,16 @@ private:
 	LineView head_line;
 	LineView msg_line;
 	LineView qna_prompt_line;
+	LineEdit qna_input;
 	std::vector<QnaPair> qna_prompts;
-	Operation finish_operation;
+	QnaFinishAction qna_finish_operation;
 	History* qna_history;
 	std::shared_ptr<FormAction> parent_formaction;
+
+	// cmdline completion
+	std::string last_fragment;
+	std::vector<std::string> suggestions;
+	unsigned int tab_count;
 };
 
 } // namespace newsboat
